import paho.mqtt.client
import asyncio
import logging
import socket
import typing

logger = logging.getLogger(__name__)


class AsyncWrapper(typing.AsyncIterable[paho.mqtt.client.MQTTMessage]):
    def __init__(self, cli: paho.mqtt.client.Client):
        self._cli = cli
        self._cli.on_socket_open = self._on_socket_open
        self._cli.on_connect = self._on_connected
        self._cli.on_socket_register_write = self._on_register_write
        self._cli.on_socket_unregister_write = self._on_unregister_write
        self._forward_messages = not self._cli.on_message
        if not self._forward_messages:
            logger.warning("on_message callback was already set for client, AsyncWrapper.get_message() will be unavailable")
        else:
            self._cli.on_message = self._on_message

        self._socket = None
        self._connected_future = asyncio.Future()
        self._mq = asyncio.Queue()
        self._want_write = False
        self._can_write = False

    def _on_register_write(self, *args):
        logger.info("Registering writer")
        asyncio.get_event_loop().add_writer(self._socket.fileno(), self._ready_to_write)

    def _on_unregister_write(self, *args):
        logger.info("Unregistering writer")
        asyncio.get_event_loop().remove_writer(self._socket.fileno())

    def _on_message(self, cli, _, msg: paho.mqtt.client.MQTTMessage):
        logger.info("Forwarding message: topic: %r, pl: %r", msg.topic, msg.payload)
        self._mq.put_nowait(msg)

    def _on_connected(self, *args):
        logger.info("Connected: %r", args)
        self._connected_future.set_result(None)
        self._cli.subscribe('/#', 2)

    def _on_socket_open(self, cli, ud, socket: socket.socket):
        logger.info("socket open: ud: %s, socket: %s", ud, socket)
        self._socket = socket
        asyncio.get_event_loop().add_reader(self._socket.fileno(), self._ready_to_read)

    def _on_socket_closed(self):
        asyncio.get_event_loop().remove_reader(self._socket.fileno())
        asyncio.get_event_loop().remove_writer(self._socket.fileno())

    def _ready_to_read(self):
        logger.debug("Ready to read")
        ret = self._cli.loop_read()
        logger.debug("loop_read() result: %r", ret)
        self._write_maybe()

    def _write_maybe(self):
        if self._cli.want_write():
            if self._can_write:
                logger.debug("Ready to write")
                self._cli.loop_write()
                self._can_write = False

    def _ready_to_write(self):
        self._can_write = True
        self._write_maybe()

    async def connect(self, *args, **kw):
        self._cli.connect(*args, **kw)
        await self._connected_future

    async def __aiter__(self):
        return self

    async def __anext__(self) -> paho.mqtt.client.MQTTMessage:
        return await self._mq.get()
